@@ -1,9 +1,10 @@
-/* api_wrdb.cc: tests which need a writable backend
- *
- * Copyright 1999,2000,2001 BrightStation PLC
+/** @file
+ * @brief tests which need a writable backend
+ */
+/* Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001 Hein Ragas
  * Copyright 2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2014,2015 Olly Betts
+ * Copyright 2002-2023 Olly Betts
  * Copyright 2006 Richard Boulton
  * Copyright 2007 Lemur Consulting Ltd
  *
@@ -29,18 +30,20 @@
 
 #include <xapian.h>
 
-#include "backendmanager.h" // For XAPIAN_BIN_PATH.
+#include "filetests.h"
 #include "omassert.h"
 #include "str.h"
+#include "stringutils.h"
 #include "testsuite.h"
 #include "testutils.h"
 #include "unixcmds.h"
-#include "utils.h"
 
 #include "apitest.h"
 
+#include "safeunistd.h"
 #include <cmath>
 #include <cstdlib>
+#include <limits>
 #include <map>
 #include <string>
 
@@ -91,12 +94,12 @@ DEFINE_TESTCASE(adddoc1, writable) {
     Xapian::MSet mset = enq.get_mset(0, 10);
 
     mset_expect_order(mset, 3, 1, 2);
-
-    return true;
 }
 
 // test that removing a posting and removing a term works
-DEFINE_TESTCASE(adddoc2, writable) {
+DEFINE_TESTCASE(adddoc2, writable && !multi) {
+    // FIXME: With multi, get_termfreq() on a TermIterator from a Document
+    // currently returns the termfreq for just the shard the doc is in.
     Xapian::WritableDatabase db = get_writable_database();
 
     Xapian::Document doc1;
@@ -123,7 +126,7 @@ DEFINE_TESTCASE(adddoc2, writable) {
     TEST_EQUAL(*iter2, *iter1);
     TEST_EQUAL(iter1.get_wdf(), 1);
     TEST_EQUAL(iter2.get_wdf(), 1);
-    //TEST_EQUAL(iter1.get_termfreq(), 0);
+    // TEST_EQUAL(iter1.get_termfreq(), 0);
     TEST_EQUAL(iter2.get_termfreq(), 1);
 
     iter1++;
@@ -134,7 +137,7 @@ DEFINE_TESTCASE(adddoc2, writable) {
     TEST_EQUAL(*iter2, *iter1);
     TEST_EQUAL(iter1.get_wdf(), 4);
     TEST_EQUAL(iter2.get_wdf(), 4);
-    //TEST_EQUAL(iter1.get_termfreq(), 0);
+    // TEST_EQUAL(iter1.get_termfreq(), 0);
     TEST_EQUAL(iter2.get_termfreq(), 1);
 
     iter1++;
@@ -177,19 +180,19 @@ DEFINE_TESTCASE(adddoc2, writable) {
     iter2 = doc2.termlist_begin();
     TEST(iter2 != doc2.termlist_end());
     TEST_EQUAL(*iter2, "bag");
-    //TEST_EQUAL(iter2.get_termfreq(), 0);
+    // TEST_EQUAL(iter2.get_termfreq(), 0);
     iter2++;
     TEST(iter2 != doc2.termlist_end());
     TEST_EQUAL(*iter2, "bar");
-    //TEST_EQUAL(iter2.get_termfreq(), 0);
+    // TEST_EQUAL(iter2.get_termfreq(), 0);
     iter2++;
     TEST(iter2 != doc2.termlist_end());
     TEST_EQUAL(*iter2, "bat");
-    //TEST_EQUAL(iter2.get_termfreq(), 0);
+    // TEST_EQUAL(iter2.get_termfreq(), 0);
     iter2++;
     TEST(iter2 != doc2.termlist_end());
     TEST_EQUAL(*iter2, "foo");
-    //TEST_EQUAL(iter2.get_termfreq(), 0);
+    // TEST_EQUAL(iter2.get_termfreq(), 0);
     iter2++;
     TEST(iter2 == doc2.termlist_end());
 
@@ -205,7 +208,7 @@ DEFINE_TESTCASE(adddoc2, writable) {
     TEST_EQUAL(iter1.get_wdf(), 0);
     TEST_EQUAL(iter2.get_wdf(), 0);
     TEST_EQUAL(iter1.get_termfreq(), 1);
-    //TEST_EQUAL(iter2.get_termfreq(), 0);
+    // TEST_EQUAL(iter2.get_termfreq(), 0);
     TEST(iter1.positionlist_begin() == iter1.positionlist_end());
     TEST(iter2.positionlist_begin() == iter2.positionlist_end());
 
@@ -218,7 +221,7 @@ DEFINE_TESTCASE(adddoc2, writable) {
     TEST_EQUAL(iter1.get_wdf(), 9);
     TEST_EQUAL(iter2.get_wdf(), 9);
     TEST_EQUAL(iter1.get_termfreq(), 2);
-    //TEST_EQUAL(iter2.get_termfreq(), 0);
+    // TEST_EQUAL(iter2.get_termfreq(), 0);
 
     Xapian::PositionIterator pi1;
     pi1 = iter1.positionlist_begin();
@@ -237,7 +240,7 @@ DEFINE_TESTCASE(adddoc2, writable) {
     TEST_EQUAL(iter1.get_wdf(), 0);
     TEST_EQUAL(iter2.get_wdf(), 0);
     TEST_EQUAL(iter1.get_termfreq(), 1);
-    //TEST_EQUAL(iter2.get_termfreq(), 0);
+    // TEST_EQUAL(iter2.get_termfreq(), 0);
     TEST(iter1.positionlist_begin() == iter1.positionlist_end());
     TEST(iter2.positionlist_begin() == iter2.positionlist_end());
 
@@ -250,7 +253,7 @@ DEFINE_TESTCASE(adddoc2, writable) {
     TEST_EQUAL(iter1.get_wdf(), 0);
     TEST_EQUAL(iter2.get_wdf(), 0);
     TEST_EQUAL(iter1.get_termfreq(), 2);
-    //TEST_EQUAL(iter2.get_termfreq(), 0);
+    // TEST_EQUAL(iter2.get_termfreq(), 0);
 
     Xapian::PositionIterator temp1 = iter1.positionlist_begin();
     pi1 = temp1;
@@ -265,8 +268,6 @@ DEFINE_TESTCASE(adddoc2, writable) {
     iter2++;
     TEST(iter1 == doc1.termlist_end());
     TEST(iter2 == doc2.termlist_end());
-
-    return true;
 }
 
 // test that adding lots of documents works, and doesn't leak memory
@@ -283,12 +284,12 @@ DEFINE_TESTCASE(adddoc3, writable) {
 	}
 	db.add_document(doc);
     }
-    return true;
 }
 
-// We want to test that a termlist starting with a 48 character long term works
-// OK since this value currently requires special handling in flint for
-// historical reasons!)  Also test all other term lengths while we're at it.
+// We originally wanted to test that a termlist starting with a 48 character
+// long term worked since that required special handling in flint for
+// historical reasons.  That's no longer relevant, but it seems useful to
+// continue to test term lists starting with various term lengths work.
 DEFINE_TESTCASE(adddoc4, writable) {
     Xapian::WritableDatabase db = get_writable_database();
 
@@ -313,13 +314,14 @@ DEFINE_TESTCASE(adddoc4, writable) {
     // And test a document with no terms.
     Xapian::Document doc = db.get_document(241);
     TEST(doc.termlist_begin() == doc.termlist_end());
-
-    return true;
 }
 
 // Test adding a document, and checking that it got added correctly.
 // This testcase used to be adddoc2 in quartztest.
-DEFINE_TESTCASE(adddoc5, writable) {
+DEFINE_TESTCASE(adddoc5, writable && !multi) {
+    // FIXME: With multi, get_termfreq() on a TermIterator from a Document
+    // currently returns the termfreq for just the shard the doc is in.
+
     // Inmemory doesn't support get_writable_database_as_database().
     SKIP_TEST_FOR_BACKEND("inmemory");
 
@@ -356,7 +358,7 @@ DEFINE_TESTCASE(adddoc5, writable) {
 	Xapian::docid did2 = database.add_document(document_in2);
 	TEST_EQUAL(database.get_doccount(), 2);
 	TEST_NOT_EQUAL(did, did2);
-	TEST_EQUAL(database.get_avlength(), 5.0/2.0);
+	TEST_EQUAL(database.get_avlength(), 5.0 / 2.0);
 
 	TEST_EQUAL(database.get_termfreq("foobar"), 2);
 	TEST_EQUAL(database.get_collection_freq("foobar"), 3);
@@ -378,7 +380,7 @@ DEFINE_TESTCASE(adddoc5, writable) {
 
 	did = database.add_document(document_in);
 	TEST_EQUAL(database.get_doccount(), 2);
-	TEST_EQUAL(database.get_avlength(), 5.0/2.0);
+	TEST_EQUAL(database.get_avlength(), 5.0 / 2.0);
 
 	TEST_EQUAL(database.get_termfreq("foobar"), 2);
 	TEST_EQUAL(database.get_collection_freq("foobar"), 3);
@@ -423,8 +425,10 @@ DEFINE_TESTCASE(adddoc5, writable) {
 		TEST_NOT_EQUAL(j, document_out.termlist_end());
 		TEST_EQUAL(*i, *j);
 		TEST_EQUAL(i.get_wdf(), j.get_wdf());
+		// Actually use termfreq to stop compiler optimising away the
+		// call to get_termfreq().
 		TEST_EXCEPTION(Xapian::InvalidOperationError,
-			       (void)i.get_termfreq());
+			       if (i.get_termfreq()) FAIL_TEST("?"));
 		TEST_NOT_EQUAL(0, j.get_termfreq());
 		if (*i == "foobar") {
 		    // termfreq of foobar is 2
@@ -445,8 +449,6 @@ DEFINE_TESTCASE(adddoc5, writable) {
 	    TEST_EQUAL(j, document_out.termlist_end());
 	}
     }
-
-    return true;
 }
 
 // Test adding a document, and checking that it got added correctly.
@@ -491,8 +493,6 @@ DEFINE_TESTCASE(adddoc6, writable) {
 	TEST_EQUAL(database.get_termfreq("baz"), 0);
 	TEST_EQUAL(database.get_collection_freq("baz"), 0);
     }
-
-    return true;
 }
 
 // tests that database destructors commit if it isn't done explicitly
@@ -506,8 +506,6 @@ DEFINE_TESTCASE(implicitendsession1, writable) {
     doc.add_posting("nsa", 2);
     doc.add_posting("fbi", 3);
     db.add_document(doc);
-
-    return true;
 }
 
 // tests that assignment of Xapian::Database and Xapian::WritableDatabase works
@@ -522,9 +520,21 @@ DEFINE_TESTCASE(databaseassign1, writable) {
     Xapian::Database d2(actually_wdb);
     d2 = wdb;
     d2 = actually_wdb;
+#ifdef __has_warning
+# if __has_warning("-Wself-assign-overloaded")
+    // Suppress warning from newer clang about self-assignment so we can
+    // test that self-assignment works!
+#  pragma clang diagnostic push
+#  pragma clang diagnostic ignored "-Wself-assign-overloaded"
+# endif
+#endif
     wdb = wdb; // check assign to itself works
     db = db; // check assign to itself works
-    return true;
+#ifdef __has_warning
+# if __has_warning("-Wself-assign-overloaded")
+#  pragma clang diagnostic pop
+# endif
+#endif
 }
 
 // tests that deletion and updating of documents works as expected
@@ -570,8 +580,6 @@ DEFINE_TESTCASE(deldoc1, writable) {
     TEST_EQUAL(*tit, "fwing");
     tit++;
     TEST_EQUAL(tit, doc2.termlist_end());
-
-    return true;
 }
 
 // tests that deletion and updating of documents works as expected
@@ -602,7 +610,8 @@ DEFINE_TESTCASE(deldoc2, writable) {
 
     db.commit();
 
-    db.reopen();
+    // reopen() on a writable database shouldn't do anything.
+    TEST(!db.reopen());
 
     db.delete_document(1);
     db.delete_document(2);
@@ -610,7 +619,8 @@ DEFINE_TESTCASE(deldoc2, writable) {
 
     db.commit();
 
-    db.reopen();
+    // reopen() on a writable database shouldn't do anything.
+    TEST(!db.reopen());
 
     TEST_EQUAL(db.postlist_begin("one"), db.postlist_end("one"));
     TEST_EQUAL(db.postlist_begin("two"), db.postlist_end("two"));
@@ -641,13 +651,15 @@ DEFINE_TESTCASE(deldoc2, writable) {
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_doclength(2));
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_doclength(3));
 
+    TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_unique_terms(1));
+    TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_unique_terms(2));
+    TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_unique_terms(3));
+
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_document(1));
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_document(2));
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_document(3));
 
     TEST_EQUAL(db.allterms_begin(), db.allterms_end());
-
-    return true;
 }
 
 // another test of deletion of documents, a cut-down version of deldoc2
@@ -663,18 +675,19 @@ DEFINE_TESTCASE(deldoc3, writable) {
 
     db.commit();
 
-    db.reopen();
+    // reopen() on a writable database shouldn't do anything.
+    TEST(!db.reopen());
 
     db.delete_document(1);
 
     db.commit();
 
-    db.reopen();
+    // reopen() on a writable database shouldn't do anything.
+    TEST(!db.reopen());
 
     TEST_EQUAL(db.postlist_begin("one"), db.postlist_end("one"));
 
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.termlist_begin(1));
-    (void)&db; // gcc 2.95 seems to miscompile without this!!! - Olly
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.termlist_begin(2));
 
     // test positionlist_{begin,end}?
@@ -690,12 +703,13 @@ DEFINE_TESTCASE(deldoc3, writable) {
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_doclength(1));
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_doclength(2));
 
+    TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_unique_terms(1));
+    TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_unique_terms(2));
+
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_document(1));
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_document(2));
 
     TEST_EQUAL(db.allterms_begin(), db.allterms_end());
-
-    return true;
 }
 
 // tests that deletion and updating of (lots of) documents works as expected
@@ -729,11 +743,13 @@ DEFINE_TESTCASE(deldoc4, writable) {
 	bool is_power_of_two = ((i & (i - 1)) == 0);
 	if (is_power_of_two) {
 	    db.commit();
-	    db.reopen();
+	    // reopen() on a writable database shouldn't do anything.
+	    TEST(!db.reopen());
 	}
     }
     db.commit();
-    db.reopen();
+    // reopen() on a writable database shouldn't do anything.
+    TEST(!db.reopen());
 
     /* delete the documents in a peculiar order */
     for (Xapian::docid i = 0; i < maxdoc / 3; ++i) {
@@ -743,8 +759,8 @@ DEFINE_TESTCASE(deldoc4, writable) {
     }
 
     db.commit();
-
-    db.reopen();
+    // reopen() on a writable database shouldn't do anything.
+    TEST(!db.reopen());
 
     TEST_EQUAL(db.postlist_begin("one"), db.postlist_end("one"));
     TEST_EQUAL(db.postlist_begin("two"), db.postlist_end("two"));
@@ -759,6 +775,7 @@ DEFINE_TESTCASE(deldoc4, writable) {
 	tout.str(string());
 	TEST_EXCEPTION(Xapian::DocNotFoundError, db.termlist_begin(i));
 	TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_doclength(i));
+	TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_unique_terms(i));
 	TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_document(i));
     }
 
@@ -779,8 +796,6 @@ DEFINE_TESTCASE(deldoc4, writable) {
     TEST_EQUAL(db.get_collection_freq("three"), 0);
 
     TEST_EQUAL(db.allterms_begin(), db.allterms_end());
-
-    return true;
 }
 
 // Test deleting a document which was added in the same batch.
@@ -826,8 +841,6 @@ DEFINE_TESTCASE(deldoc5, writable) {
     TEST_EQUAL(*p, 3);
     ++p;
     TEST_EQUAL(p, db.postlist_end("foo"));
-
-    return true;
 }
 
 // Regression test for bug in quartz and flint, fixed in 1.0.2.
@@ -857,8 +870,6 @@ DEFINE_TESTCASE(deldoc6, writable) {
     db.commit();
 
     TEST_EXCEPTION(Xapian::DocNotFoundError, db.get_document(2));
-
-    return true;
 }
 
 DEFINE_TESTCASE(replacedoc1, writable) {
@@ -868,7 +879,7 @@ DEFINE_TESTCASE(replacedoc1, writable) {
 
     doc1.add_posting("foo", 1);
     doc1.add_posting("foo", 2);
-    doc1.add_posting("gone",3);
+    doc1.add_posting("gone", 3);
     doc1.add_posting("bar", 4);
     doc1.add_posting("foo", 5);
     Xapian::docid did;
@@ -886,23 +897,22 @@ DEFINE_TESTCASE(replacedoc1, writable) {
     db.replace_document(did, doc2);
 
     Xapian::Document doc3 = db.get_document(did);
-    Xapian::TermIterator tIter = doc3.termlist_begin();
-    TEST_EQUAL(*tIter, "bar");
-    Xapian::PositionIterator pIter = tIter.positionlist_begin();
-    TEST_EQUAL(*pIter, 4);
-    ++tIter;
-    TEST_EQUAL(*tIter, "foo");
-    Xapian::PositionIterator qIter = tIter.positionlist_begin();
-    TEST_EQUAL(*qIter, 1);
-    ++qIter;
-    TEST_EQUAL(*qIter, 5);
-    ++tIter;
-    TEST_EQUAL(*tIter, "pipco");
-    Xapian::PositionIterator rIter = tIter.positionlist_begin();
-    TEST_EQUAL(*rIter, 2);
-    ++tIter;
-    TEST_EQUAL(tIter, doc3.termlist_end());
-    return true;
+    Xapian::TermIterator t_iter = doc3.termlist_begin();
+    TEST_EQUAL(*t_iter, "bar");
+    Xapian::PositionIterator p_iter = t_iter.positionlist_begin();
+    TEST_EQUAL(*p_iter, 4);
+    ++t_iter;
+    TEST_EQUAL(*t_iter, "foo");
+    Xapian::PositionIterator q_iter = t_iter.positionlist_begin();
+    TEST_EQUAL(*q_iter, 1);
+    ++q_iter;
+    TEST_EQUAL(*q_iter, 5);
+    ++t_iter;
+    TEST_EQUAL(*t_iter, "pipco");
+    Xapian::PositionIterator r_iter = t_iter.positionlist_begin();
+    TEST_EQUAL(*r_iter, 2);
+    ++t_iter;
+    TEST_EQUAL(t_iter, doc3.termlist_end());
 }
 
 // Test of new feature: WritableDatabase::replace_document accepts a docid
@@ -914,7 +924,7 @@ DEFINE_TESTCASE(replacedoc2, writable) {
 
     doc1.add_posting("foo", 1);
     doc1.add_posting("foo", 2);
-    doc1.add_posting("gone",3);
+    doc1.add_posting("gone", 3);
     doc1.add_posting("bar", 4);
     doc1.add_posting("foo", 5);
     Xapian::docid did = 31770;
@@ -938,30 +948,28 @@ DEFINE_TESTCASE(replacedoc2, writable) {
     TEST_EQUAL(db.get_doccount(), 1);
 
     Xapian::Document doc3 = db.get_document(did);
-    Xapian::TermIterator tIter = doc3.termlist_begin();
-    TEST_EQUAL(*tIter, "bar");
-    Xapian::PositionIterator pIter = tIter.positionlist_begin();
-    TEST_EQUAL(*pIter, 4);
-    ++tIter;
-    TEST_EQUAL(*tIter, "foo");
-    Xapian::PositionIterator qIter = tIter.positionlist_begin();
-    TEST_EQUAL(*qIter, 1);
-    ++qIter;
-    TEST_EQUAL(*qIter, 5);
-    ++tIter;
-    TEST_EQUAL(*tIter, "pipco");
-    Xapian::PositionIterator rIter = tIter.positionlist_begin();
-    TEST_EQUAL(*rIter, 2);
-    ++tIter;
-    TEST_EQUAL(tIter, doc3.termlist_end());
+    Xapian::TermIterator t_iter = doc3.termlist_begin();
+    TEST_EQUAL(*t_iter, "bar");
+    Xapian::PositionIterator p_iter = t_iter.positionlist_begin();
+    TEST_EQUAL(*p_iter, 4);
+    ++t_iter;
+    TEST_EQUAL(*t_iter, "foo");
+    Xapian::PositionIterator q_iter = t_iter.positionlist_begin();
+    TEST_EQUAL(*q_iter, 1);
+    ++q_iter;
+    TEST_EQUAL(*q_iter, 5);
+    ++t_iter;
+    TEST_EQUAL(*t_iter, "pipco");
+    Xapian::PositionIterator r_iter = t_iter.positionlist_begin();
+    TEST_EQUAL(*r_iter, 2);
+    ++t_iter;
+    TEST_EQUAL(t_iter, doc3.termlist_end());
 
     did = db.add_document(doc1);
     TEST_EQUAL(did, 31771);
     TEST_EQUAL(db.get_doccount(), 2);
 
     TEST_EXCEPTION(Xapian::InvalidArgumentError, db.replace_document(0, doc2));
-
-    return true;
 }
 
 // Test replacing a document which was added in the same batch.
@@ -1005,14 +1013,20 @@ DEFINE_TESTCASE(replacedoc3, writable) {
     TEST_EQUAL(db.get_doclength(2), 1);
     TEST_EQUAL(db.get_doclength(3), 4);
 
+    TEST_EQUAL(db.get_unique_terms(1), 3);
+    TEST_EQUAL(db.get_unique_terms(2), 1);
+    TEST_EQUAL(db.get_unique_terms(3), 4);
+
     Xapian::PostingIterator p = db.postlist_begin("foo");
     TEST_NOT_EQUAL(p, db.postlist_end("foo"));
     TEST_EQUAL(*p, 1);
     TEST_EQUAL(p.get_doclength(), 3);
+    TEST_EQUAL(p.get_unique_terms(), 3);
     ++p;
     TEST_NOT_EQUAL(p, db.postlist_end("foo"));
     TEST_EQUAL(*p, 3);
     TEST_EQUAL(p.get_doclength(), 4);
+    TEST_EQUAL(p.get_unique_terms(), 4);
     ++p;
     TEST_EQUAL(p, db.postlist_end("foo"));
 
@@ -1020,14 +1034,14 @@ DEFINE_TESTCASE(replacedoc3, writable) {
     TEST_NOT_EQUAL(p, db.postlist_end("world"));
     TEST_EQUAL(*p, 2);
     TEST_EQUAL(p.get_doclength(), 1);
+    TEST_EQUAL(p.get_unique_terms(), 1);
     ++p;
     TEST_NOT_EQUAL(p, db.postlist_end("world"));
     TEST_EQUAL(*p, 3);
     TEST_EQUAL(p.get_doclength(), 4);
+    TEST_EQUAL(p.get_unique_terms(), 4);
     ++p;
     TEST_EQUAL(p, db.postlist_end("world"));
-
-    return true;
 }
 
 // Test replacing a document which was deleted in the same batch.
@@ -1086,13 +1100,11 @@ DEFINE_TESTCASE(replacedoc4, writable) {
     TEST_EQUAL(*p, 3);
     ++p;
     TEST_EQUAL(p, db.postlist_end("world"));
-
-    return true;
 }
 
 // Test replacing a document with itself without modifying postings.
 // Regression test for bug in 0.9.9 and earlier - there flint and quartz
-// lose all positional information for the document when you do this.
+// lost all positional information for the document when you did this.
 DEFINE_TESTCASE(replacedoc5, writable) {
     Xapian::WritableDatabase db = get_writable_database();
 
@@ -1119,9 +1131,9 @@ DEFINE_TESTCASE(replacedoc5, writable) {
 	TEST(db.positionlist_begin(1, "world") != db.positionlist_end(1, "world"));
     }
 
-    // Brass, chert and flint now spot simple cases of replacing the same
-    // document and don't do needless work.  Force them to actually do the
-    // replacement to make sure that case works.
+    // The backends now spot simple cases of replacing the same document and
+    // don't do needless work.  Force them to actually do the replacement to
+    // make sure that case works.
 
     {
 	Xapian::Document doc;
@@ -1143,8 +1155,6 @@ DEFINE_TESTCASE(replacedoc5, writable) {
 	TEST(db.positionlist_begin(1, "hello") != db.positionlist_end(1, "hello"));
 	TEST(db.positionlist_begin(1, "world") != db.positionlist_end(1, "world"));
     }
-
-    return true;
 }
 
 // Test replacing a document while adding values, without changing anything
@@ -1183,8 +1193,6 @@ DEFINE_TESTCASE(replacedoc6, writable) {
     doc = db.get_document(1);
     TEST_EQUAL(doc.get_value(1), "banana1");
     TEST_EQUAL(doc.get_value(2), "banana2");
-
-    return true;
 }
 
 // Test of new feature: WritableDatabase::replace_document and delete_document
@@ -1239,8 +1247,6 @@ DEFINE_TESTCASE(uniqueterm1, writable) {
     db.delete_document("U2");
 
     TEST_EQUAL(db.get_doccount(), 16);
-
-    return true;
 }
 
 // tests all document postlists
@@ -1282,8 +1288,6 @@ DEFINE_TESTCASE(allpostlist2, writable) {
 	}
     }
     TEST_EQUAL(j, 512);
-
-    return true;
 }
 
 static void test_emptyterm2_helper(Xapian::WritableDatabase & db)
@@ -1327,29 +1331,29 @@ DEFINE_TESTCASE(emptyterm2, writable) {
 	TEST_EQUAL(db.get_doccount(), 0);
 	test_emptyterm2_helper(db);
     }
-
-    return true;
 }
 
 // Check that PHRASE/NEAR becomes AND if there's no positional info in the
 // database.
-DEFINE_TESTCASE(phraseorneartoand1, writable) {
-    Xapian::WritableDatabase db = get_writable_database();
-
-    for (int n = 1; n <= 20; ++n) {
-	Xapian::Document doc;
-	doc.add_term(str(n));
-	doc.add_term(str(n ^ 9));
-	doc.add_term("all");
-	doc.set_data("pass1");
-	db.add_document(doc);
-    }
-    db.commit();
+DEFINE_TESTCASE(phraseorneartoand1, backend) {
+    Xapian::Database db = get_database("phraseorneartoand1",
+				       [](Xapian::WritableDatabase& wdb,
+					  const string&)
+				       {
+					   for (int n = 1; n <= 20; ++n) {
+					       Xapian::Document doc;
+					       doc.add_term(str(n));
+					       doc.add_term(str(n ^ 9));
+					       doc.add_term("all");
+					       doc.set_data("pass1");
+					       wdb.add_document(doc);
+					   }
+				       });
 
     Xapian::Enquire enquire(db);
     Xapian::MSet mymset;
 
-    const char * q1[] = { "all", "1" };
+    static const char * const q1[] = { "all", "1" };
     enquire.set_query(Xapian::Query(Xapian::Query::OP_PHRASE, q1, q1 + 2));
     mymset = enquire.get_mset(0, 10);
     TEST_EQUAL(2, mymset.size());
@@ -1358,7 +1362,7 @@ DEFINE_TESTCASE(phraseorneartoand1, writable) {
     mymset = enquire.get_mset(0, 10);
     TEST_EQUAL(2, mymset.size());
 
-    const char * q2[] = { "1", "2" };
+    static const char * const q2[] = { "1", "2" };
     enquire.set_query(Xapian::Query(Xapian::Query::OP_PHRASE, q2, q2 + 2));
     mymset = enquire.get_mset(0, 10);
     TEST_EQUAL(0, mymset.size());
@@ -1366,27 +1370,35 @@ DEFINE_TESTCASE(phraseorneartoand1, writable) {
     enquire.set_query(Xapian::Query(Xapian::Query::OP_NEAR, q2, q2 + 2));
     mymset = enquire.get_mset(0, 10);
     TEST_EQUAL(0, mymset.size());
+}
 
-    return true;
+static void
+gen_longpositionlist1_db(Xapian::WritableDatabase& db, const string&)
+{
+    Xapian::Document doc;
+    for (Xapian::termpos n = 1; n <= 2000; ++n) {
+	doc.add_posting("fork", n * 3);
+	doc.add_posting("knife", n * unsigned(log(double(n + 2))));
+	doc.add_posting("spoon", n * n);
+	// Exercise positions up to 4 billion.
+	Xapian::termpos half_cube = n * n / 2 * n;
+	doc.add_posting("chopsticks", half_cube);
+	if (sizeof(Xapian::termpos) >= 8) {
+	    // Exercise 64-bit positions.
+	    doc.add_posting("spork", half_cube * half_cube);
+	}
+    }
+    doc.set_data("cutlery");
+    db.add_document(doc);
 }
 
 // Check that a large number of position list entries for a particular term
 // works - regression test for flint.
-DEFINE_TESTCASE(longpositionlist1, writable) {
-    Xapian::WritableDatabase db = get_writable_database();
+DEFINE_TESTCASE(longpositionlist1, backend) {
+    Xapian::Database db = get_database("longpositionlist1",
+				       gen_longpositionlist1_db);
 
-    Xapian::Document doc;
-    Xapian::termpos n;
-    for (n = 1; n <= 2000; ++n) {
-	doc.add_posting("fork", n * 3);
-	doc.add_posting("knife", n * unsigned(log(double(n + 2))));
-	doc.add_posting("spoon", n * n);
-    }
-    doc.set_data("cutlery");
-    Xapian::docid did = db.add_document(doc);
-    db.commit();
-
-    doc = db.get_document(did);
+    Xapian::Document doc = db.get_document(1);
 
     Xapian::TermIterator t, tend;
     Xapian::PositionIterator p, pend;
@@ -1394,11 +1406,24 @@ DEFINE_TESTCASE(longpositionlist1, writable) {
     t = doc.termlist_begin();
     tend = doc.termlist_end();
 
+    TEST(t != tend);
+    TEST_EQUAL(*t, "chopsticks");
+    p = t.positionlist_begin();
+    pend = t.positionlist_end();
+    for (Xapian::termpos n = 1; n <= 2000; ++n) {
+	TEST(p != pend);
+	Xapian::termpos half_cube = n * n / 2 * n;
+	TEST_EQUAL(*p, half_cube);
+	++p;
+    }
+    TEST(p == pend);
+
+    ++t;
     TEST(t != tend);
     TEST_EQUAL(*t, "fork");
     p = t.positionlist_begin();
     pend = t.positionlist_end();
-    for (n = 1; n <= 2000; ++n) {
+    for (Xapian::termpos n = 1; n <= 2000; ++n) {
 	TEST(p != pend);
 	TEST_EQUAL(*p, n * 3);
 	++p;
@@ -1410,7 +1435,7 @@ DEFINE_TESTCASE(longpositionlist1, writable) {
     TEST_EQUAL(*t, "knife");
     p = t.positionlist_begin();
     pend = t.positionlist_end();
-    for (n = 1; n <= 2000; ++n) {
+    for (Xapian::termpos n = 1; n <= 2000; ++n) {
 	TEST(p != pend);
 	TEST_EQUAL(*p, n * unsigned(log(double(n + 2))));
 	++p;
@@ -1422,28 +1447,39 @@ DEFINE_TESTCASE(longpositionlist1, writable) {
     TEST_EQUAL(*t, "spoon");
     p = t.positionlist_begin();
     pend = t.positionlist_end();
-    for (n = 1; n <= 2000; ++n) {
+    for (Xapian::termpos n = 1; n <= 2000; ++n) {
 	TEST(p != pend);
 	TEST_EQUAL(*p, n * n);
 	++p;
     }
     TEST(p == pend);
 
+    if (sizeof(Xapian::termpos) >= 8) {
+	++t;
+	TEST(t != tend);
+	TEST_EQUAL(*t, "spork");
+	p = t.positionlist_begin();
+	pend = t.positionlist_end();
+	for (Xapian::termpos n = 1; n <= 2000; ++n) {
+	    TEST(p != pend);
+	    Xapian::termpos half_cube = n * n / 2 * n;
+	    TEST_EQUAL(*p, half_cube * half_cube);
+	    ++p;
+	}
+	TEST(p == pend);
+    }
+
     ++t;
     TEST(t == tend);
-
-    return true;
 }
 
-// Regression test for bug#110: Inconsistent sort order between pages with
-// set_sort_by_value_then_relevance.
-DEFINE_TESTCASE(consistency2, writable) {
-    Xapian::WritableDatabase db = get_writable_database();
+static void
+gen_consistency2_db(Xapian::WritableDatabase& db, const string&)
+{
     char buf[2] = "X";
-    int i = 0;
 
     // Add 5 documents indexed by "test" with wdf 1.
-    for (i = 0; i < 5; ++i) {
+    for (int i = 0; i < 5; ++i) {
 	Xapian::Document doc;
 	*buf = '0' + i;
 	doc.add_value(0, buf);
@@ -1452,15 +1488,19 @@ DEFINE_TESTCASE(consistency2, writable) {
     }
 
     // Add 5 documents indexed by "test" with wdf 2.
-    for (i = 0; i < 5; ++i) {
+    for (int i = 0; i < 5; ++i) {
 	Xapian::Document doc;
 	*buf = '0' + i;
 	doc.add_value(0, buf);
 	doc.add_term("test", 2);
 	db.add_document(doc);
     }
+}
 
-    db.commit();
+// Regression test for bug#110: Inconsistent sort order between pages with
+// set_sort_by_value_then_relevance.
+DEFINE_TESTCASE(consistency2, backend) {
+    Xapian::Database db = get_database("consistency2", gen_consistency2_db);
 
     Xapian::Enquire enq(db);
     enq.set_query(Xapian::Query("test"));
@@ -1481,14 +1521,15 @@ DEFINE_TESTCASE(consistency2, writable) {
 
     TEST_EQUAL(*mset1[0], *mset2a[0]);
     TEST_EQUAL(*mset1[1], *mset2b[0]);
-    for (i = 0; i < 8; ++i) {
+    for (Xapian::doccount i = 0; i < 8; ++i) {
 	TEST_EQUAL(*mset1[i + 2], *mset2c[i]);
     }
-
-    return true;
 }
 
-DEFINE_TESTCASE(crashrecovery1, brass || chert || flint) {
+DEFINE_TESTCASE(crashrecovery1, chert) {
+    // Glass has a single version file per revision, rather than multiple base
+    // files, so it simply can't get into the situations we are testing
+    // recovery from.
     const string & dbtype = get_dbtype();
     string path = ".";
     path += dbtype;
@@ -1511,7 +1552,7 @@ DEFINE_TESTCASE(crashrecovery1, brass || chert || flint) {
 
 	db.add_document(doc);
 	db.commit();
-	dbr.reopen();
+	TEST(dbr.reopen());
 	TEST_EQUAL(dbr.get_doccount(), 1);
 
 	// Xapian::Database has full set of baseB, old baseA
@@ -1524,7 +1565,7 @@ DEFINE_TESTCASE(crashrecovery1, brass || chert || flint) {
 
 	db.add_document(doc);
 	db.commit();
-	dbr.reopen();
+	TEST(dbr.reopen());
 	TEST_EQUAL(dbr.get_doccount(), 2);
 
 	// Xapian::Database has full set of baseA, old baseB
@@ -1537,10 +1578,10 @@ DEFINE_TESTCASE(crashrecovery1, brass || chert || flint) {
 
 	// Simulate a transaction starting, some of the baseB getting removed,
 	// but then the transaction fails.
-	unlink(path + "/record" + base_ext);
-	unlink(path + "/termlist" + base_ext);
+	unlink((path + "/record" + base_ext).c_str());
+	unlink((path + "/termlist" + base_ext).c_str());
 
-	dbr.reopen();
+	TEST(!dbr.reopen());
 	TEST_EQUAL(dbr.get_doccount(), 2);
     }
 
@@ -1556,7 +1597,7 @@ DEFINE_TESTCASE(crashrecovery1, brass || chert || flint) {
 
     db.add_document(doc);
     db.commit();
-    dbr.reopen();
+    TEST(dbr.reopen());
     TEST_EQUAL(dbr.get_doccount(), 3);
 
     // Xapian::Database has full set of baseB, old baseA
@@ -1569,17 +1610,16 @@ DEFINE_TESTCASE(crashrecovery1, brass || chert || flint) {
 
     db.add_document(doc);
     db.commit();
-    dbr.reopen();
+    TEST(dbr.reopen());
     TEST_EQUAL(dbr.get_doccount(), 4);
-
-    return true;
 }
 
 // Check that DatabaseError is thrown if the docid counter would wrap.
 // Regression test for bug#152.
 DEFINE_TESTCASE(nomoredocids1, writable) {
     // The InMemory backend uses a vector for the documents, so trying to add
-    // document "-1" will fail because we can't allocate enough memory!
+    // a document with a really large docid will fail because we can't allocate
+    // enough memory!
     SKIP_TEST_FOR_BACKEND("inmemory");
 
     Xapian::WritableDatabase db = get_writable_database();
@@ -1587,11 +1627,10 @@ DEFINE_TESTCASE(nomoredocids1, writable) {
     doc.set_data("prose");
     doc.add_term("word");
 
-    db.replace_document(Xapian::docid(-1), doc);
+    Xapian::docid max_id = numeric_limits<Xapian::docid>::max();
+    db.replace_document(max_id, doc);
 
     TEST_EXCEPTION(Xapian::DatabaseError, db.add_document(doc));
-
-    return true;
 }
 
 // Test synonym iterators.
@@ -1704,8 +1743,6 @@ DEFINE_TESTCASE(synonymitor1, writable && synonyms) {
 	s += '|';
     }
     TEST_STRINGS_EQUAL(s, "|hello|");
-
-    return true;
 }
 
 // Test that adding a document with a really long term gives an error on
@@ -1716,9 +1753,9 @@ DEFINE_TESTCASE(termtoolong1, writable) {
 
     Xapian::WritableDatabase db = get_writable_database();
 
-    for (Xapian::doccount i = 246; i <= 290; ++i) {
+    for (size_t i = 246; i <= 290; ++i) {
 	tout.str(string());
-	tout << "Term length " << i << endl;
+	tout << "Term length " << i << '\n';
 	Xapian::Document doc;
 	string term(i, 'X');
 	doc.add_term(term);
@@ -1729,14 +1766,14 @@ DEFINE_TESTCASE(termtoolong1, writable) {
 	    // Check that the max length is correctly expressed in the
 	    // exception message - we've got this wrong in two different ways
 	    // in the past!
-	    tout << e.get_msg() << endl;
+	    tout << e.get_msg() << '\n';
 	    TEST(e.get_msg().find("Term too long (> 245)") != string::npos);
 	}
     }
 
-    for (Xapian::doccount j = 240; j <= 245; ++j) {
+    for (size_t j = 240; j <= 245; ++j) {
 	tout.str(string());
-	tout << "Term length " << j << endl;
+	tout << "Term length " << j << '\n';
 	Xapian::Document doc;
 	string term(j, 'X');
 	doc.add_term(term);
@@ -1745,12 +1782,13 @@ DEFINE_TESTCASE(termtoolong1, writable) {
 
     db.commit();
 
+    size_t limit = endswith(get_dbtype(), "glass") ? 255 : 252;
     {
-	// Currently brass, flint and chert escape zero bytes from terms in
-	// keys for some tables, so a term with 127 zero bytes won't work
-	// either.
+	// Currently chert and glass escape zero bytes from terms in keys for
+	// some tables, so a term with 127 zero bytes won't work for chert, and
+	// with 128 zero bytes won't work for glass.
 	Xapian::Document doc;
-	doc.add_term(string(127, '\0'));
+	doc.add_term(string(limit / 2 + 1, '\0'));
 	db.add_document(doc);
 	try {
 	    db.commit();
@@ -1759,12 +1797,26 @@ DEFINE_TESTCASE(termtoolong1, writable) {
 	    // Check that the max length is correctly expressed in the
 	    // exception message - we've got this wrong in two different ways
 	    // in the past!
-	    tout << e.get_msg() << endl;
-	    TEST(e.get_msg().find(" is 252 bytes") != string::npos);
+	    tout << e.get_msg() << '\n';
+	    string target = " is ";
+	    target += str(limit);
+	    target += " bytes";
+	    TEST(e.get_msg().find(target) != string::npos);
 	}
     }
 
-    return true;
+    if (get_dbtype().find("chert") != string::npos) {
+	XFAIL("Chert fails to clear pending changes after "
+	      "InvalidArgumentError - fix too invasive");
+    }
+    // Try adding a document.  Regression test for a bug fixed in 1.4.15 - in
+    // earlier versions the pending changes which caused the
+    // InvalidArgumentError were left around and a subsequent commit() on the
+    // same WritableDatabase would also fail with InvalidArgumentError.
+    Xapian::Document doc;
+    doc.add_term("ok");
+    db.add_document(doc);
+    db.commit();
 }
 
 /// Test playing with a postlist
@@ -1784,6 +1836,7 @@ DEFINE_TESTCASE(postlist7, writable) {
     TEST_EQUAL(*p, 5);
     TEST_EQUAL(p.get_wdf(), 3);
     TEST_EQUAL(p.get_doclength(), 7);
+    TEST_EQUAL(p.get_unique_terms(), 2);
     ++p;
     TEST(p == db_w.postlist_end("foo"));
 
@@ -1799,50 +1852,46 @@ DEFINE_TESTCASE(postlist7, writable) {
     TEST_EQUAL(*p, 5);
     TEST_EQUAL(p.get_wdf(), 3);
     TEST_EQUAL(p.get_doclength(), 7);
+    TEST_EQUAL(p.get_unique_terms(), 2);
     ++p;
     TEST(p != db_w.postlist_end("foo"));
     TEST_EQUAL(*p, 6);
     TEST_EQUAL(p.get_wdf(), 1);
     TEST_EQUAL(p.get_doclength(), 2);
+    TEST_EQUAL(p.get_unique_terms(), 2);
     ++p;
     TEST(p == db_w.postlist_end("foo"));
-
-    return true;
 }
 
-DEFINE_TESTCASE(lazytablebug1, brass || chert || flint) {
-    {
-	Xapian::WritableDatabase db = get_named_writable_database("lazytablebug1", string());
+static void
+gen_lazytablebug1_db(Xapian::WritableDatabase& db, const string&)
+{
+    Xapian::Document doc;
+    doc.add_term("foo");
+    db.add_document(doc);
+    db.commit();
 
-	Xapian::Document doc;
-	doc.add_term("foo");
-	db.add_document(doc);
-	db.commit();
-
-	string synonym(255, 'x');
-	char buf[] = " iamafish!!!!!!!!!!";
-	for (int i = 33; i < 120; ++i) {
-	    db.add_synonym(buf, synonym);
-	    ++buf[0];
-	}
-
-	db.commit();
+    string synonym(255, 'x');
+    char buf[] = " iamafish!!!!!!!!!!";
+    for (int i = 33; i < 120; ++i) {
+	db.add_synonym(buf, synonym);
+	++buf[0];
     }
+}
 
-    Xapian::Database db = get_writable_database_as_database();
+DEFINE_TESTCASE(lazytablebug1, synonyms) {
+    Xapian::Database db = get_database("lazytablebug1", gen_lazytablebug1_db);
     for (Xapian::TermIterator t = db.synonym_keys_begin(); t != db.synonym_keys_end(); ++t) {
-	tout << *t << endl;
+	tout << *t << '\n';
     }
-
-    return true;
 }
 
 /** Regression test for bug #287 for flint.
  *
- *  Chert also has the same duff code but this testcase doesn't actually 
+ *  Chert also has the same duff code but this testcase doesn't actually
  *  tickle the bug there.
  */
-DEFINE_TESTCASE(cursordelbug1, brass || chert || flint) {
+DEFINE_TESTCASE(cursordelbug1, writable && path) {
     static const int terms[] = { 219, 221, 222, 223, 224, 225, 226 };
     static const int copies[] = { 74, 116, 199, 21, 45, 155, 189 };
 
@@ -1868,17 +1917,8 @@ DEFINE_TESTCASE(cursordelbug1, brass || chert || flint) {
 
     db.commit();
 
-    string cmd = XAPIAN_BIN_PATH"xapian-check ";
-    cmd += get_named_writable_database_path("cursordelbug1");
-#ifdef __WIN32__
-    cmd += " >nul";
-#else
-    cmd += " >/dev/null";
-#endif
-    if (system(cmd.c_str()) != 0)
-	return false;
-
-    return true;
+    const string & db_path = get_named_writable_database_path("cursordelbug1");
+    TEST_EQUAL(Xapian::Database::check(db_path), 0);
 }
 
 /** Helper function for modifyvalues1.
@@ -1888,7 +1928,7 @@ static void
 check_vals(const Xapian::Database & db, const map<Xapian::docid, string> & vals)
 {
     TEST_EQUAL(db.get_doccount(), vals.size());
-    if (vals.size() == 0) return;
+    if (vals.empty()) return;
     TEST_REL(vals.rbegin()->first,<=,db.get_lastdocid());
     map<Xapian::docid, string>::const_iterator i;
     for (i = vals.begin(); i != vals.end(); ++i) {
@@ -1920,13 +1960,13 @@ DEFINE_TESTCASE(modifyvalues1, writable) {
     Xapian::WritableDatabase db = get_writable_database();
     // Note: doccount must be coprime with 13
     const Xapian::doccount doccount = 1000;
-    STATIC_ASSERT(doccount % 13 != 0);
+    static_assert(doccount % 13 != 0, "doccount divisible by 13");
 
     map<Xapian::docid, string> vals;
 
     for (Xapian::doccount num = 1; num <= doccount; ++num) {
 	tout.str(string());
-    	Xapian::Document doc;
+	Xapian::Document doc;
 	string val = "val" + str(num);
 	tout << "Setting val '" << val << "' in doc " << num << "\n";
 	doc.add_value(1, val);
@@ -2012,6 +2052,33 @@ DEFINE_TESTCASE(modifyvalues1, writable) {
     check_vals(db, vals);
     db.commit();
     check_vals(db, vals);
+}
 
-    return true;
+/** Regression test for protocol design bug.
+ *
+ *  Previously some messages didn't send a reply but could result in an
+ *  exception being sent over the link.  That exception would then get
+ *  read as a response to the next message instead of its actual response
+ *  so we'd be out of step.
+ *
+ *  This also affected MSG_DELETEDOCUMENTTERM, MSG_CANCEL, MSG_SETMETADATA
+ *  and MSG_ADDSPELLING but it's harder to reliably trigger an exception
+ *  from any of those.
+ *
+ *  See #783.  Fixed in 1.4.12.
+ */
+DEFINE_TESTCASE(protocolbug1, remote && writable) {
+    Xapian::WritableDatabase db = get_writable_database("");
+    Xapian::Document doc;
+    doc.add_term(string(300, 'x'));
+
+    TEST_EXCEPTION(Xapian::InvalidArgumentError,
+		   db.replace_document(1, doc));
+    db.commit();
+}
+
+DEFINE_TESTCASE(remotefdleak1, remote && writable) {
+    Xapian::WritableDatabase wdb = get_writable_database("");
+    TEST_EXCEPTION(Xapian::DatabaseLockError,
+		   auto wdb2 = get_writable_database_again());
 }
